"""Dialogs for the UI."""
from kivy.uix.floatlayout import FloatLayout
from kivy.factory import Factory
from kivy.properties import ObjectProperty


class LoadDialog(FloatLayout):

    """A dialog to open files."""

    load = ObjectProperty(None)
    cancel = ObjectProperty(None)


class SaveDialog(FloatLayout):

    """A dialog to save files."""

    save = ObjectProperty(None)
    text_input = ObjectProperty(None)
    cancel = ObjectProperty(None)

Factory.register('LoadDialog', cls=LoadDialog)
Factory.register('SaveDialog', cls=SaveDialog)

__all__ = ["LoadDialog", "SaveDialog"]



"""Observable Translations for kivy that change when the language changes.

The functionality of this module is highly inspired by
`kivy-gettext-example <https://github.com/tito/kivy-gettext-example>`.

"""
from kivy.lang import Observable


class ObservableTranslation(Observable):

    """This class allows kivy translations to be updated with the language."""

    def __init__(self, translate):
        """Create a new translation object with a translation function.

        :param translate: a callable that translates the text. Even when the
          language is changed, it returns the text for the currrent language.
        """
        super().__init__()
        self._translate = translate
        self._observers = []

    def __call__(self, text):
        """Call this object to translate text.

        :param str text: the text to translate
        :return: the text translated to the current language
        """
        return self._translate(text)

    def fbind(self, name, func, args, **kwargs):
        """Add an observer. This is used by kivy."""
        self._observers.append((name, func, args, kwargs))

    def funbind(self, name, func, args, **kwargs):
        """Remove an observer. This is used by kivy."""
        key = (name, func, args, kwargs)
        if key in self._observers:
            self._observers.remove(key)

    def language_changed(self):
        """Update all the kv rules attached to this text."""
        for name, func, args, kwargs in self._observers:
            func(args, None, None)

__all__ = ["ObservableTranslation"]



from kivy.factory import Factory
from kivy.uix.textinput import TextInput
from kivy.properties import NumericProperty
import re

_make_integer_pattern = re.compile('[^0-9]')


def make_integer(string):
    """Extract the interger from a string.

    :param str string: the string to make to an integer
    :return: the string as an integer
    :rtype: str
    """
    return _make_integer_pattern.sub("", string)


class IntInput(TextInput):

    """An input that only takes integers.

    .. seealso:: :mod:`kivy.uix.textinput`
    """

    #: the value as an integer
    value = NumericProperty(0)

    def insert_text(self, substring, from_undo=False):
        """When text is inserted."""
        string = make_integer(substring)
        return super().insert_text(string, from_undo=from_undo)

    def on_text(self, instance, text):
        """When the text changes, change the :attr:`value`."""
        if str(self.value) != text:
            self.value = int(text or 0)

    def on_value(self, instance, value):
        """when the :attr:`value` changes, change the text."""
        if self.text != str(int(value)):
            self.text = str(int(value))


Factory.register('IntInput', cls=IntInput)

__all__ = ["make_integer", "IntInput"]



import test_localization
from kniteditor.localization import _, change_language_to, list_languages, \
    DEFAULT_LANGUAGE, current_language, language_code_to_translation, \
    translation_to_language_code, list_translated_languages, \
    current_translated_language, change_language_to_translated
import pytest
from pytest import raises


def test_current_language_is_default_language():
    assert current_language() == DEFAULT_LANGUAGE


def test_default_language_is_english():
    assert DEFAULT_LANGUAGE == "en"


@pytest.mark.parametrize("language", ["de", "en"])
def test_languages_are_listed(language):
    assert language in list_languages()


@pytest.mark.parametrize("word,translation", [("yes", "yes"), ("no", "no")])
def test_translate_in_default_language(word, translation):
    assert _(word) == translation


@pytest.mark.parametrize("language,word,translation", [
    ("en", "Yes", "Yes"), ("en", "No", "No"),
    ("de", "Yes", "Ja"), ("de", "No", "Nein")])
def test_translate_in_default_language(language, word, translation):
    change_language_to(language)
    assert _(word) == translation


@pytest.mark.parametrize("language", ["de", "en", "en", "de", "de"])
def test_switching_language_sets_current_language(language):
    change_language_to(language)
    assert current_language() == language


@pytest.mark.parametrize("code,translation", [("de", "Deutsch"),
                                              ("en", "English")])
def test_language_translation(code, translation):
    assert language_code_to_translation(code) == translation
    assert translation_to_language_code(translation) == code


@pytest.mark.parametrize("code", ["invalid!!!!", "123"])
def test_value_error_in_language_code_to_translation(code):
    with raises(ValueError) as error:
        language_code_to_translation(code)
    message = "Invalid language code {}. Expected one of {}.".format(
        repr(code), ", ".join(map(repr, list_languages())))
    assert error.value.args[0] == message


@pytest.mark.parametrize("code", ["invalid!!!!", "123", "en"])
@pytest.mark.parametrize("function", [change_language_to_translated,
                                      translation_to_language_code])
def test_value_error_in_translation_to_language_code(code, function):
    with raises(ValueError) as error:
        function(code)
    message = "Invalid language name {}. Expected one of {}.".format(
        repr(code), ", ".join(map(repr, map(language_code_to_translation,
                                            list_languages()))))
    assert error.value.args[0] == message


@pytest.mark.parametrize("code", [None, object(), object])
@pytest.mark.parametrize("function", [language_code_to_translation,
                                      translation_to_language_code])
def test_type_error_in_language_code_to_translation(code, function):
    with raises(TypeError) as error:
        function(code)
    message = "Invalid argument {}. Expected str instance.".format(
        repr(code))
    assert error.value.args[0] == message


def test_list_translated_languages_have_no_duplicate():
    translated_languages = list_translated_languages()
    assert len(set(translated_languages)) == len(translated_languages)


@pytest.mark.parametrize("code", list_languages())
def test_all_languages_in_translated_languages(code):
    assert language_code_to_translation(code) in list_translated_languages()


@pytest.mark.parametrize("code", list_languages())
def test_can_set_translated_language(code):
    translated = language_code_to_translation(code)
    change_language_to_translated(translated)
    assert current_language() == code
    assert current_translated_language() == translated



"""This module provides translation functionality.

The function to use for translation is :class:`_(string)
<kniteditor.localization.observable_translation.ObservableTranslation>`
to translate strings. When you use this in the .kv file, the translationss
are automatically updated when the language is changed.


"""
import os
import gettext
from os.path import abspath, join, dirname
from .observable_translation import ObservableTranslation

_here = dirname(__file__)
_locale_dir = abspath(join(_here, 'translations'))
_current_language = None
_locales = None
DOMAIN = "kniteditor"  #: the name of the .po files

_languages_dir = "languages"
_languages_locales = gettext.translation(_languages_dir, _locale_dir,
                                         languages=[_languages_dir])


def _(string):
    """Translate a string using the current language.

    :param str string: the string to translate
    :return: the translated string
    :rtype: str
    """
    return _locales.gettext(string)
_ = ObservableTranslation(_)


def change_language_to(new_language):
    """Change the language to a language folder in the translations folder.

    :param str new_language: The language code to translate everything into.
      This code must be listed in :func:`list_languages`.
    :raises ValueError: if :paramref:`new_language` is not listed by
      :func:`list_languages`
    """
    assert new_language in list_languages()
    global _locales, _current_language
    _locales = gettext.translation(DOMAIN, _locale_dir,
                                   languages=[new_language])
    _current_language = new_language
    _.language_changed()


_supported_languages = []
for folder_name in os.listdir(_locale_dir):
    folder = os.path.join(_locale_dir, folder_name)
    if folder_name != _languages_dir and os.path.isdir(folder):
        if "LC_MESSAGES" in os.listdir(folder):
            _supported_languages.append(folder_name)
    del folder, folder_name


def list_languages():
    """Return a list of all supported languages.

    :return: a list of :class:`strings <str>` of language codes that can be
      passed as an argument to :func:`change_language_to`
    :rtype: list
    """
    return _supported_languages.copy()


def language_code_to_translation(language_code):
    """Translate a language code.

    Note that all language names are written in their language so that
    people can read them in their language.

    :rtype: str
    :return: the translated language
    :param str language_code: a language code from :func:`list_languages`
    :raises ValueError: if the code is not in :func:`list_languages`
    :raises TypeError: if the code is not a string

    .. note:: this operation can be reversed by
      :func:`translation_to_language_code`

    .. code:: python

        c2 = language_code_to_translation(language_code_to_translation(c1))
        assert c1 == c2

    """
    if not isinstance(language_code, str):
        message = "Invalid argument {}. Expected str instance.".format(
            repr(language_code))
        raise TypeError(message)
    if language_code not in list_languages():
        message = "Invalid language code {}. Expected one of {}.".format(
            repr(language_code), ", ".join(map(repr, list_languages())))
        raise ValueError(message)
    return _languages_locales.gettext(language_code)


def translation_to_language_code(translated_language_code):
    """Translate a language back to a langugage code.

    :param str translated_language_code: a language returned by
      :func:`language_code_to_translation`
    :rtype: str
    :return: a language code in :func:`list_languages`
    :raises ValueError: if this translation is not known
    """
    if not isinstance(translated_language_code, str):
        message = "Invalid argument {}. Expected str instance.".format(
            repr(translated_language_code))
        raise TypeError(message)
    expected = []
    for language_code in list_languages():
        translation = language_code_to_translation(language_code)
        expected.append(translation)
        if translation == translated_language_code:
            return language_code
    message = "Invalid language name {}. Expected one of {}.".format(
        repr(translated_language_code), ", ".join(map(repr, expected)))
    raise ValueError(message)


def current_language():
    """Return the current language.

    :return: a language listed by :func:`list_languages`
    :rtype: str
    """
    return _current_language


def list_translated_languages():
    """Return a list of translated language names.

    :rtype: list
    :return: a list of :func:`translated language names
      <language_code_to_translation>`
    """
    return list(map(language_code_to_translation, list_languages()))


def current_translated_language():
    """The current language as a translated string.

    :rtype: str
    :return: the :func:`translated <language_code_to_translation>` version of
      the :func:`current language <current_language>`
    """
    return language_code_to_translation(current_language())


def change_language_to_translated(new_language):
    """Set the language to a translated language name.

    :param str new_language: a language form :func:`list_translated_languages`
    :raises ValueError: if :paramref:`new_language` is not listed in
      :func:`list_translated_languages`
    """
    new_language_code = translation_to_language_code(new_language)
    assert new_language_code in list_languages()
    change_language_to(new_language_code)


DEFAULT_LANGUAGE = "en"  #: the default language to use
change_language_to(DEFAULT_LANGUAGE)

__all__ = ["_", "change_language_to", "list_languages", "DEFAULT_LANGUAGE",
           "current_language",
           "language_code_to_translation", "translation_to_language_code",
           "change_language_to_translated", "current_translated_language",
           "list_translated_languages"]



"""This module contains the editor window."""
import knittingpattern
import os
import sys
from kivy.app import App
from kivy.uix.pagelayout import PageLayout
from kivy.uix.popup import Popup
from kivy.properties import ObjectProperty
from .dialogs import LoadDialog, SaveDialog
from .localization import _, list_translated_languages, \
    change_language_to_translated, current_translated_language
import json

LANGUAGE_CODE = "current"  #: the language code name
LANGUAGE_SECTION = "language"  #: the language section name


class Root(PageLayout):

    """This is the root of the application."""

    knitting_pattern = ObjectProperty(None)
    knit_settings = ObjectProperty(None)

    def show_open_file_dialog(self):
        """Open the file dialog for loading."""
        content = LoadDialog(load=self.load_path, cancel=self.dismiss_popup)
        self._popup = Popup(title="Load file", content=content,
                            size_hint=(0.9, 0.9))
        self._popup.open()

    def show_save_file_dialog(self):
        """Open the file dialog for saving."""
        content = SaveDialog(save=self.save_path, cancel=self.dismiss_popup)
        self._popup = Popup(title="Save file", content=content,
                            size_hint=(0.9, 0.9))
        self._popup.open()

    def load_path(self, path, filenames):
        """Load knitting patterns from several files."""
        file_path = filenames[0]
        extension = os.path.splitext(file_path.lower())[1]
        if extension == ".json":
            patterns = knittingpattern.load_from_path(file_path)
        else:
            converter = knittingpattern.convert_from_image()
            patterns = converter.path(file_path).knitting_pattern()
        pattern = patterns.patterns.at(0)
        self.knitting_pattern.show_pattern(pattern)
        self.dismiss_popup()

    def save_path(self, path, filename):
        """Save a knitting pattern to a path."""
        print("save")

        self.dismiss_popup()

    def dismiss_popup(self):
        """Close the popup window."""
        self._popup.dismiss()


class EditorWindow(App):

    """The editor window."""

    def get_application_config(self):
        """Return the application's configuration directory.

        :return: the path to the configuration file
        :rtype: str

        .. seealso:: :meth:`kivy.app.App.get_application_config`
        """
        return super().get_application_config('~/.%(appname)s.ini')

    def build_config(self, config):
        """Build the configuration.

        :param kivy.config.ConfigParser config: the configuration parser

        .. seealso:: `Application Configuration
         <https://kivy.org/docs/api-kivy.app.html#application-configuration>`__
        """
        config.setdefaults(LANGUAGE_SECTION, {
            LANGUAGE_CODE: current_translated_language()
        })

    def build_settings(self, settings):
        """Create the applications settings dialog.

        :param  kivy.uix.settings.Settings settings: the settings for this app

        .. seealso:: `Create a settings panel
          <https://kivy.org/docs/api-kivy.app.html#create-a-settings-panel>`__,
          :meth:`kivy.uix.settings.Settings.add_json_panel`,
          :mod:`kivy.uix.settings`
        """
        settings.add_json_panel(_('KnitEditor'), self.config,
                                data=self.settings_specification)

    def update_language_from_config(self):
        """Set the current language of the application from the configuration.
        """
        config_language = self.config.get(LANGUAGE_SECTION, LANGUAGE_CODE)
        change_language_to_translated(config_language)

    @property
    def settings_specification(self):
        """The settings specification as JSON string.

        :rtype: str
        :return: a JSON string


        """
        settings = [
            {"type": "title",
             "title": _("KnitEditor")},

            {"type": "options",
             "title": _("Language"),
             "desc": _("Choose your language"),
             "section": "language",
             "key": LANGUAGE_CODE,
             "options": list_translated_languages()},
        ]
        return json.dumps(settings)

    def on_config_change(self, config, section, key, value):
        """The configuration was changed.

        :param kivy.config.ConfigParser config: the configuration that was
          changed
        :param str section: the section that was changed
        :param str key: the key in the section that was changed
        :param value: the value this key was changed to

        When this method is called, it issued calls to change methods if they
        exist in this order:

        - ``config_change_in_section_{section}_key_{key}(value)``
        - ``config_change_in_section_{section}(key, value)``
        """
        section_call = "config_change_in_section_{}".format(section)
        key_call = "{}_key_{}".format(section_call, key)
        if hasattr(self, key_call):
            getattr(self, key_call)(value)
        elif hasattr(self, section_call):
            getattr(self, section_call)(key, value)

    def config_change_in_section_language_key_current(self, new_language):
        """Set the new language of the application.

        Same as :func:`kniteditor.localization.change_language_to_translated`
        """
        change_language_to_translated(new_language)

    def build(self):
        """Build the application."""
        self.update_language_from_config()

    def on_start(self):
        """The application is started.

        .. seealso:: :meth:`kivy.app.App.on_start`
        """
        self.show_example()

    def on_stop(self):
        """The application terminates.

        .. seealso:: :meth:`kivy.app.App.on_stop`
        """
        self.root.knit_settings.stop()

    def show_example(self):
        """Show an example knitting pattern."""
        patterns = knittingpattern.load_from().example("block4x4.json")
        pattern = patterns.patterns.at(0)
        self.root.knitting_pattern.show_pattern(pattern)


def main(argv=sys.argv):
    """Open the editor window."""
    if "/test" in argv:
        import pytest
        errcode = pytest.main(["--pyargs", "knittingpattern", "kniteditor",
                               "AYABInterface"])
        sys.exit(errcode)
    EditorWindow().run()

__all__ = ["main", "EditorWindow", "Root", "LANGUAGE_CODE", "LANGUAGE_SECTION"]



"""A dialog for the settings to knit."""
from kivy.factory import Factory
from kivy.uix.boxlayout import BoxLayout
from kivy.properties import StringProperty, ObjectProperty
from kivy.uix.button import Button
from kivy.uix.popup import Popup
from AYABInterface import get_machines, get_connections
from AYABInterface.interaction import Interaction
from AYABInterface.machines import KH910
from kivy.logger import Logger
from .localization import _
from kivy.uix.dropdown import DropDown


class NullCommunication(object):

    """A null object pattern for communication."""

    def stop(self):
        """Do nothing."""


class NullFile(object):

    """A null object pattern for files."""

    def close(self):
        """Do nothing."""

FIRST_MACHINE = KH910()  #: the machine that is displayed as default


class DebugSerial(object):

    """A :class:`serial.Serial` interface proxy that logs the communication."""

    def __init__(self, serial):
        """Create a new DebugSerial object.

        :param serial.Serial serial: the serial interface
        """
        self._serial = serial
        self.close = serial.close

    def write(self, bytes_):
        """Write bytes to the serial and log them.

        :param bytes bytes_: the bytes to write
        """
        Logger.debug("write:", bytes_)
        self._serial.write(bytes_)

    def read(self, *args):
        """Read bytes from the serial and log them.

        :param tuple args: the arguments to pass on to the read method
        :return: the :class:`bytes` read from the serial
        :rtype: bytes
        """
        bytes_ = self._serial.read(*args)
        Logger.debug("read:", bytes_)
        return bytes_


class AYABKnitSettings(BoxLayout):

    """Class containing the settings to connect to the AYAB shield."""

    #: the different colors chosen
    colors_layout = ObjectProperty(None)

    # common

    def populate_drop_down(self, entries, attribute, location):
        """Populate a drop down menu."""
        drop_down = DropDown()
        for entry in entries:
            Logger.info("populate dropdown for {}: {}"
                        "".format(attribute, entry))
            button = Button(
                text=entry.name, height=44, size_hint_y=None,
                on_release=lambda a, entry=entry:
                    setattr(self, attribute, entry))
            drop_down.add_widget(button)
        drop_down.open(location)

    def display_message(self, title, message):
        """Create a pop up error message.

        .. seealso:: :mod:`kivu.uix.popup`
        """
        if self.popup:
            self.popup.dismiss()
        content = Button(text=message, on_press=lambda i: popup.dismiss())
        popup = Popup(title=title, content=content,
                      size_hint=(0.5, 0.5))
        self.popup = popup
        popup.open()

    #: the :class:`~kivy.uix.popup.Popup` to display messages to the user
    popup = ObjectProperty(None)

    # the machines

    #: the machine to use for knitting
    machine = ObjectProperty(FIRST_MACHINE)
    #: the name of the :attr:`machine`
    machine_name = StringProperty(FIRST_MACHINE.name)
    #: the button to choose the machines.
    machine_button = ObjectProperty(None)

    def populate_machines_drop_down(self):
        """Add elements to the machines choice."""
        self.populate_drop_down(get_machines(), "machine", self.machine_button)

    def on_machine(self, instance, value):
        """The machine changed so we need to adapt the machine text."""
        Logger.info("machine chosen: {}".format(value))
        self.machine_name = value.name

    # the connections

    #: the connection to communicate through
    connection = ObjectProperty(None)
    #: the name of the connection
    connection_name = StringProperty(_("Choose a connection!"))
    #: the button to choose the connection with
    connection_button = ObjectProperty(None)

    def populate_connections_drop_down(self):
        """Add elements to the connection choice."""
        self.populate_drop_down(get_connections(),
                                "connection", self.connection_button)

    def on_connection(self, instance, value):
        """The connection changed so we need to adapt the machine text."""
        Logger.info("connection chosen: {}".format(value))
        self.connection_name = value.name

    # knitting

    def start_knitting(self, pattern):
        """Start the knitting process with one pattern."""
        if self.connection is None:
            self.display_message(
                _("Please Configure!"),
                _("Please choose a connection before knitting."))
            return
        self.pattern_in_progress.show_pattern(pattern)
        self.start_knitting_button.text = _("Restart knitting!")

        self.stop()

        self.interaction = Interaction(pattern, self.machine)
        self.comunication_connection = DebugSerial(self.connection.connect())
        self.communication = self.interaction.communicate_through(
            self.comunication_connection)
        self.communication.on_message(print)
        self.communication.parallelize(2)

        self.populate_actions(self.interaction.actions)

    def stop(self):
        """Stop the connections."""
        self.comunication_connection.close()
        self.communication.stop()

    def build(self):
        """Build the UI for the first time."""
        self.ayab_connections_dropdown.dismiss()
        self.ayab_machines_dropdown.dismiss()

    def populate_actions(self, actions):
        """Show the actions associted with the pattern."""
        self.list_of_actions.clear_widgets()
        for action in actions:
            button = Button(text=str(action))
            self.list_of_actions.add_widget(button)

    #: the :class:`~AYABInterface.interaction.Interaction` object to
    #: interact with while knitting
    interaction = ObjectProperty(None)
    #: the :class:`~AYABInterface.communication.Communication` object to
    #: communicate with
    communication = ObjectProperty(NullCommunication())
    #: the :class:`DebugSerial` or :class:`serial.Serial` to use for
    #: communication
    comunication_connection = ObjectProperty(NullFile())
    #: the list of actions to diaplay to the user
    list_of_actions = ObjectProperty(None)
    #: the :class:`~kniteditor.KnittingPatternWidget.KnittingPatternWidget` to
    #: display the pattern to knit and the progess
    pattern_in_progress = ObjectProperty(None)
    #: the :class:`~kivy.uix.button.Button` to start and restart knitting with
    start_knitting_button = ObjectProperty(None)

Factory.register('AYABKnitSettings', cls=AYABKnitSettings)
__all__ = ["AYABKnitSettings", "DebugSerial", "NullCommunication", "NullFile"]
